#include "board.h"
#include "fs_tools.h"


FATFS fs;				/* File system object */
DIR dir;				/* Directory object */
FIL fileR;				/* File object */
FIL fileW;				/* File object */
FILINFO fileInfo;		// file/dir info object
static char resTextBuffer[RES_TEXT_BUFFER_SIZE]; //buffer for result text description

//internal funcs
FRESULT fsSetSize(FIL *fp, uint32_t size);
FRESULT fsDecreaseSize(FIL *fp, uint32_t bytes);
FRESULT fsFindValueForward(FIL *fp, uint8_t byte);


//=================================================================================================
//								HIGH-LEVEL UTILS (USE FILE NAME)
//=================================================================================================

void fsListFiles(char *dirPath)
{
	FILINFO info;
	DIR currentDir;
	
	if(f_opendir(&currentDir, dirPath) == FR_OK)
	{
		while ((f_readdir(&currentDir, &info) == FR_OK) && info.fname[0])
		{
			debug_msg(info.fname);
		}
	}
}


FRESULT fsRemove(char* path)
{
	FRESULT res = f_unlink((const char*)path);
	if(res!=FR_OK)
	{
		debug_msg("ERR fsRemove error");
		fsDispFRESULT(res,resTextBuffer);
	}
	return(res);
}



FRESULT fsMkDir(char* dirPath)
{
	FRESULT res = f_mkdir((const char*)dirPath);
	if(res!=FR_OK)
	{
		debug_msg("ERR fsMkDir error");
		fsDispFRESULT(res,resTextBuffer);
	}
	return(res);
}



FRESULT fsChDir(char* dirPath)
{
	FRESULT res = f_opendir(&dir,(const char*)dirPath);
	if(res!=FR_OK)
	{
		debug_msg("ERR fsChDir error");
		fsDispFRESULT(res,resTextBuffer);
	}
	return(res);
}



FRESULT fsGetFileSize(char* fileName, uint32_t *size)
{
	FRESULT res = f_stat((const char*)fileName,&fileInfo);
	if(res!=FR_OK)
	{
		*size = 0xFFFFFFFF;
		fsDispFRESULT(res,resTextBuffer);
		return(res);
	}
	*size = fileInfo.fsize;
	return(res);
}


FRESULT fsRewrite(char* fileName, void *buf, uint16_t len)
{
	//open file
	FRESULT res = f_open(&fileW,(const char*)fileName,FA_WRITE|FA_CREATE_ALWAYS);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRewrite: f_open error");
		fsDispFRESULT(res,resTextBuffer);
		f_close(&fileW);
		return(res);
	}
	//set cursor in file
	res = f_lseek(&fileW,0);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRewrite: f_lseek error");
		fsDispFRESULT(res,resTextBuffer);
		f_close(&fileW);
		return(res);
	}
	//write buffer
	uint16_t written=0;
	res = f_write(&fileW,buf,len,&written);
	if((res != FR_OK)||(written!=len))
	{
		debug_msg("ERR fsRewrite: f_write error");
		fsDispFRESULT(res,resTextBuffer);
		debug_value("expected len =",len);
		debug_value("bytes actually written =",written);
		f_close(&fileW);
		if(written!=len) return(FR_EOF_REACHED);
		return(res);
	}
	//close file
	res = f_close(&fileW);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRewrite: f_close error");
		fsDispFRESULT(res,resTextBuffer);
		return(res);
	}
	
	return(res);
}


//returns number of bytes read
//fileName: a pointer to file name string
//buf: pointer to buffer for data from file
//pos: starting byte in file
//len: number of bytes to read from file
uint16_t fsRead(char* fileName, void *buf, uint32_t pos, uint16_t len)
{
	//open file
	FRESULT res = f_open(&fileR,(const char*)fileName,FA_READ);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_open error");
		fsDispFRESULT(res,resTextBuffer);
		f_close(&fileR);
		return(0);
	}
	//set cursor in file
	res = f_lseek(&fileR,pos);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_lseek error");
		fsDispFRESULT(res,resTextBuffer);
		f_close(&fileR);
		return(0);
	}
	//read buffer
	uint16_t bytesRead=0;
	res = f_read(&fileR,buf,len,&bytesRead);
	
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_read error");
		fsDispFRESULT(res,resTextBuffer);
		f_close(&fileR);
		return(0);
	}
	
	if(bytesRead!=len)
	{
		debug_msg("WRN fsRead: f_read warning");
		fsDispFRESULT(res,resTextBuffer);
		debug_value("expected len =",len);
		debug_value("bytes actually read =",bytesRead);
		f_close(&fileR);
		return(bytesRead);
	}
	
	//close file
	res = f_close(&fileR);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRead: f_close error");
		fsDispFRESULT(res,resTextBuffer);
	}
	
	return(bytesRead);
}


FRESULT fsRemoveBytes(char* fileName, uint32_t bytes)
{
	FRESULT res = f_open(&fileW,(const char*)fileName,FA_WRITE|FA_OPEN_ALWAYS);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRemoveBytes: f_open error");
		fsDispFRESULT(res,resTextBuffer);
		return(res);
	}
	//delete number of bytes
	if(bytes <= (fileW.fsize)) fileW.fsize -= bytes; else fileW.fsize = 0;
	//set flag
	fileW.flag |= FA__WRITTEN;
	//close file
	res = f_close(&fileW);
	if(res != FR_OK)
	{
		debug_msg("ERR fsRemoveBytes: f_close error");
		fsDispFRESULT(res,resTextBuffer);
		return(res);
	}
	
	return(res);
}


inline FRESULT fsWriteLine(char* fileName, char *buf)
{
	FRESULT res;
	res = fsAppend((char*)fileName,buf,strlen((char*)buf));
	if(res != FR_OK) return(res);
	res = fsAppend(fileName,(uint8_t*)"\r\n",2);
	return(res);
}


inline FRESULT fsAppend(char* fileName, void *buf, uint16_t len)
{
	FRESULT res = fsReplace(fileName,buf,F_LSEEK_EOF,len);
	if(res != FR_OK)
	{
		debug_msg("ERR fsAppend: fsReplaceBufInFile error");
		fsDispFRESULT(res,resTextBuffer);
		return(res);
	}
	return(res);
}


FRESULT fsReplace(char* fileName, void *buf, uint32_t pos, uint16_t len)
{
	//open file
	FRESULT res = f_open(&fileW,(const char*)fileName,FA_WRITE|FA_OPEN_ALWAYS);
	if(res != FR_OK)
	{
		debug_msg("ERR fsReplaceBufInFile: f_open error");
		fsDispFRESULT(res,resTextBuffer);
		return(res);
	}
	//set cursor in file
	res = f_lseek(&fileW,pos);
	if(res != FR_OK)
	{
		debug_msg("ERR fsReplaceBufInFile: f_lseek error");
		fsDispFRESULT(res,resTextBuffer);
		f_close(&fileW);
		return(res);
	}
	//write buffer
	uint16_t written=0;
	res = f_write(&fileW,buf,len,&written);
	if((res != FR_OK)||(written!=len))
	{
		debug_msg("ERR fsReplaceBufInFile: f_write error");
		fsDispFRESULT(res,resTextBuffer);
		debug_value("expected len =",len);
		debug_value("bytes actually written =",written);
		f_close(&fileW);
		if(written!=len) return(FR_EOF_REACHED);
		return(res);
	}
	//close file
	res = f_close(&fileW);
	if(res != FR_OK)
	{
		debug_msg("ERR fsReplaceBufInFile: f_close error");
		fsDispFRESULT(res,resTextBuffer);
		return(res);
	}
	
	return(res);
}





//=================================================================================================
//								DEBUG UTILS
//=================================================================================================


void fsDispFRESULT(FRESULT res, char *buf)
{
	debug_msg("FRESULT = ");
	switch(res)
	{
		case FR_OK: strcpy(buf, "FR_OK"); break;
		case FR_NOT_READY: strcpy(buf, "FR_NOT_READY"); break;
		case FR_NO_FILE: strcpy(buf, "FR_NO_FILE"); break;
		case FR_NO_PATH: strcpy(buf, "FR_NO_PATH"); break;
		case FR_INVALID_NAME: strcpy(buf, "FR_INVALID_NAME"); break;
		case FR_DENIED: strcpy(buf, "FR_DENIED"); break;
		case FR_DISK_FULL: strcpy(buf, "FR_DISK_FULL"); break;
		case FR_RW_ERROR: strcpy(buf, "FR_RW_ERROR"); break;
		case FR_INCORRECT_DISK_CHANGE: strcpy(buf, "FR_INCORRECT_DISK_CHANGE"); break;
		case FR_WRITE_PROTECTED: strcpy(buf, "FR_WRITE_PROTECTED"); break;
		case FR_NOT_ENABLED: strcpy(buf, "FR_NOT_ENABLED"); break;
		case FR_NO_FILESYSTEM: strcpy(buf, "FR_NO_FILESYSTEM"); break;
		case FR_EOF_REACHED: strcpy(buf, "FR_EOF_REACHED"); break;
		default: strcpy(buf, "!!! UNDEFINED CODE");
	}
	debug_txt((char*)buf);
}


//=================================================================================================
//								INITIALIZATION UTILS
//=================================================================================================

FRESULT fsInit(void)
{
	/* Enable FatFs module */
	memset(&fs, 0, sizeof(FATFS));
	FatFs = &fs;
	
	debug_msg("ff f_mountdrv...");
	FRESULT mountRetval=f_mountdrv();
	fsDispFRESULT(mountRetval,resTextBuffer);
	return(mountRetval);
}


