#include "common.h"
#include "spi.h"
#include "lcd3310.h"


#include "lcd/bindefs.h"
#include "lcd/font2.h"
#include "lcd/font3.h"



#define LCD_PIO             AT91C_BASE_PIOA
#define LCD_SPI             AT91C_BASE_SPI0
#define LCD_DC_MASK         (1<<11)
#define LCD_RST_MASK        (1<<10)
#define LCD_NPCS_PIO        AT91C_BASE_PIOA
#define LCD_NPCS_MASK       (1<<13)
#define LCD_NPCS_PERIPH     0
#define LCD_NPCS_CODE       1
#define LCD_NPCS            1




#define LCD_DC_1            LCD_PIO->PIO_SODR = LCD_DC_MASK
#define LCD_DC_0            LCD_PIO->PIO_CODR = LCD_DC_MASK
#define LCD_RST_1           LCD_PIO->PIO_SODR = LCD_RST_MASK
#define LCD_RST_0           LCD_PIO->PIO_CODR = LCD_RST_MASK


static uint8_t lcdShadow[LCD_X_SIZE][LCD_Y_SIZE_LINES];


volatile uint8_t Font=FONT_BOLD;
volatile uint8_t TextBackground=0;


static inline void lcdSPIWrite(uint8_t value);



void lcdSetFont(uint8_t paramFont)
{
	Font = paramFont;
}

uint8_t lcdGetFont(void)
{
	return(Font);
}

void lcdBackgroundOn(void)
{
	TextBackground = 1;
}

void lcdBackgroundOff(void)
{
	TextBackground = 0;
}

uint8_t lcdClear(void)
{
	uint8_t i,j;
	for(i=0;i<LCD_X_SIZE;i++)
	{
		for(j=0;j<LCD_Y_SIZE_LINES;j++)
		{
			lcdShadow[i][j]=0;
		}
	}
	return(0);
}






uint8_t lcdInvertArea(uint8_t starting_x, uint8_t ending_x, uint8_t starting_y, uint8_t ending_y)
{
    //value boundaries
    if((ending_y-starting_y)<8) return(0);
    uint8_t upper_base_column=starting_y/8;
    uint8_t lower_base_column=ending_y/8;
    uint8_t upper_shift=starting_y%8;
    uint8_t lower_shift=ending_y%8;
    uint8_t x,y,backup_cell_positive,backup_cell_negative,i;
    for(x=starting_x;x<=ending_x;x++)
    {
        for(y=upper_base_column;y<=lower_base_column;y++)
        {
            if(y==upper_base_column)
            {
                backup_cell_positive=lcdShadow[x][y];
                backup_cell_negative=~backup_cell_positive;
                for(i=0;i<8;i++)
                {
                    if(i<=upper_shift)
                    {
                        backup_cell_negative&=~(1<<i);
                    }
                    else
                    {
                        backup_cell_positive&=~(1<<i);
                    }
                }
                lcdShadow[x][y]=backup_cell_negative|backup_cell_positive;
                continue;
            }//if(y==upper_base_column)
            if(y==lower_base_column)
            {
                backup_cell_positive=lcdShadow[x][y];
                backup_cell_negative=~backup_cell_positive;
                for(i=0;i<8;i++)
                {
                    if(i<=lower_shift)
                    {
                        backup_cell_positive&=~(1<<i);
                    }
                    else
                    {
                        backup_cell_negative&=~(1<<i);
                    }
                }
                lcdShadow[x][y]=backup_cell_negative|backup_cell_positive;
                continue;
            }//if(y==lower_base_column)
            lcdShadow[x][y]=~lcdShadow[x][y];
        }//for(y=upper_base_column;y<=lower_base_column;y++)
    }//for(x=starting_x;x<=ending_x;x++)
    return(1);
}


//===================================================================
//Text/Char Display Funcs
//===================================================================



uint8_t lcd(const char *text, uint8_t x, uint8_t y)
{
    while(*text) x=lcdChar(*text++,x,y);
    return(x);
}





//returns x value of displayed char
uint8_t lcdChar(uint8_t chr, uint8_t x, uint8_t y)
{
    uint16_t temp;
    uint8_t end_iter,i;
    uint8_t temp1;
    temp1 = font_2_ascii[chr];
    temp = temp1 * FONT_2_CHAR_SIZE;

    switch(Font)
	{
		case FONT_BOLD:
		{
			end_iter = font2[temp];
			for(i=0;i<end_iter;i++)
			{
				temp++;
				temp1 = font2[temp];
				lcdWrite(temp1,x,y);
				x++;
			}
			lcdWrite(0,x,y);
			x++;
			return(x);
		}
		case FONT_NORMAL:
		{
			end_iter = font3[temp];
			for(i=0;i<end_iter;i++)
			{
				temp++;
				temp1 = font3[temp];
				lcdWrite(temp1,x,y);
				x++;
			}
			lcdWrite(0,x,y);
			x++;
			return(x);
		}
		default:
			return(x);
	}
}





uint8_t lcdPixelSet(uint8_t x, uint8_t y)
{
	if(x>LCD_X_SIZE) return(0);
	if(y>LCD_Y_SIZE) return(0);
	uint8_t base_row=y/8;
	uint8_t pix_in_column = y%8;
	lcdShadow[x][base_row]|=(1<<pix_in_column);
	return(1);
}


uint8_t lcdPixelClear(uint8_t x, uint8_t y)
{
	if(x>LCD_X_SIZE) return(0);
	if(y>LCD_Y_SIZE) return(0);
	uint8_t base_row=y/8;
	uint8_t pix_in_column = y%8;
	lcdShadow[x][base_row]&=~(1<<pix_in_column);
	return(1);
}



uint8_t lcdPixelToggle(uint8_t x, uint8_t y)
{
	if(x>LCD_X_SIZE) return(0);
	if(y>LCD_Y_SIZE) return(0);
	uint8_t base_row=y/8;
	uint8_t pix_in_column = y%8;
	if(lcdShadow[x][base_row]&(1<<pix_in_column))
		lcdShadow[x][base_row]&=~(1<<pix_in_column);
	else
		lcdShadow[x][base_row]|=(1<<pix_in_column);
	return(1);
}



uint8_t lcdWrite(uint8_t data, uint8_t x, uint8_t y)
{
    //value boundaries
    y++;
    if(x>=LCD_X_SIZE) return(0);
    if(y>(LCD_Y_SIZE-8)) return(0);
    //prepare exact address
    uint8_t base_row=y/8;
    uint8_t hi_shift=y%8;
    uint8_t lo_shift=8-hi_shift;
    //split input data into 2 variables
    uint8_t temp_lo=(data>>lo_shift);
    uint8_t temp_hi=(data<<hi_shift);
    if(TextBackground)
    {
        //prepare clearing fields
        uint8_t clear_lo=~(0xFF>>lo_shift);
        uint8_t clear_hi=~(0xFF<<hi_shift);
        //clear shadow area
        lcdShadow[x][base_row+1]&=clear_lo;
        lcdShadow[x][base_row]&=clear_hi;
    }
    //write to prepared shadow area
    lcdShadow[x][base_row+1]|=temp_lo;
    lcdShadow[x][base_row]|=temp_hi;
    return(1);
}




//====================== THIS CODE IS DEVICE SPECIFIC ===========================


void lcdRefresh(void)
{
	LCD_DC_0;
	//y=0
	lcdSPIWrite( 0x40 );
	//x=0
	lcdSPIWrite( 0x80 );
	LCD_DC_1;
	
	spiDmaTransfer((void*)lcdShadow, LCD_X_SIZE*LCD_Y_SIZE_LINES, LCD_NPCS);
	
}


static void lcdDelay(void)
{
    uint16_t i;
    for(i=0;i<0xFFFF;i++)
    {
        asm volatile("nop"::);
    }
    //vTaskDelay(1);
}


static inline void lcdSPIWrite(uint8_t value)
{
	spiTransfer(value,LCD_NPCS);
}


void lcdContrast(uint8_t contrast)
{
    const uint8_t LCD_CONTRAST_MAX=88;
    const uint8_t LCD_CONTRAST_MIN=30;
	if(contrast>LCD_CONTRAST_MAX) contrast=LCD_CONTRAST_MAX;
	if(contrast<LCD_CONTRAST_MIN) contrast=LCD_CONTRAST_MIN;
	contrast = contrast + 128;
	LCD_DC_0;
	lcdSPIWrite(0x21); //extended command set
	lcdSPIWrite(contrast);
	lcdSPIWrite(0x22); //basic command set
	LCD_DC_1;

}



void lcdInit(void)
{
	//lcd port init
	LCD_PIO->PIO_PER = LCD_DC_MASK|LCD_RST_MASK;
	LCD_PIO->PIO_CODR = LCD_DC_MASK|LCD_RST_MASK;
	LCD_PIO->PIO_OER = LCD_DC_MASK|LCD_RST_MASK;
	//lcd SPI init
	const uint8_t SPI_SPCK_DIVIDER = 6;
	const uint32_t SCBR_MASK = (uint32_t)SPI_SPCK_DIVIDER << 8;
	LCD_SPI->SPI_CSR[LCD_NPCS] = AT91C_SPI_BITS_8 | SCBR_MASK | AT91C_SPI_NCPHA;
	LCD_NPCS_PIO->PIO_PDR = LCD_NPCS_MASK; //disable pio on NPCS pin
	//lcd NPCS pin init
	#if LCD_PERIPH==0
		LCD_NPCS_PIO->PIO_ASR=LCD_NPCS_MASK; //connected to periph A of PIO controller
	#else
		LCD_NPCS_PIO->PIO_BSR=LCD_NPCS_MASK; //connected to periph A of PIO controller
	#endif
	//lcd chip init
	//    lcdDelay();
	LCD_RST_0;
	lcdDelay();
	LCD_RST_1;
	lcdDelay();
	LCD_DC_0;
	lcdSPIWrite(0x21); //H=1 <- Extended instruction set
	lcdSPIWrite(0x30); //Vop setting
	lcdSPIWrite(0x22); //H=0 <- Basic Instruction set
	lcdSPIWrite(0x0C); //Display Config: Normal Mode
	lcdSPIWrite(0x40); //Set Y-Address of RAM (=0)
	lcdSPIWrite(0x80); //Set X-Address of RAM (=0)
	lcdSPIWrite(0x0C); //normal mode
	LCD_DC_1;
	
	uint16_t i,j;
	
	//initialize shadow
	for(j=0;j<LCD_Y_SIZE_LINES;j++)
	{
		for(i=0;i<LCD_X_SIZE;i++)
		{
			lcdShadow[i][j]=0x00;
		}
	}
	
	lcdRefresh();
}




