#include "Board.h"
#include "common.h"
#include "dbgu.h"
#include "spi.h"
#include "sd_spi.h"
#include "fat/fs_tools.h"



void init(void)
{
	// enable peripheral clock for PIO controllers
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA);
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
	
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	//open debug channel
	debug_init_default();
	
	//initialize LEDs
	LED_R_INIT;
	LED_Y_INIT;
	LED_G_INIT;
	LED_B_INIT;
	
	//initialize SPI interface
	spiInit();
	
	//initialize SD Card
	if(sdCardInit() == OK)
	{
		debug_msg("SD card initialized OK");
	}
	else
	{
		debug_msg("SD card initialization FAILED!");
	}
	
	//mount filesystem
	fsInit();
	
}

//from root
//#define FILE_NAME_LEN		(8+1+3+1) //8 chars name, 1 char dot, 3 chars extention, 1 char NULL

//longer paths
#define FILE_NAME_LEN		(50)

int main( void )
{
	init();
	
	for(;;)
	{
		uint8_t key;
		key = debug_inkey();
		switch(key)
		{
			case 'W':
			{
				debug_msg("Stream write to file. Enter file name: ");
				char fileName[FILE_NAME_LEN];
				debug_input_string(fileName, FILE_NAME_LEN);
				
				//create file structure
				FIL file;
				
				//open file for writing
				//create file if requested file doesn't exist
				f_open(
					&file,						//a pointer to file
					(const char*)fileName,		//pointer to filename
					FA_WRITE | FA_CREATE_ALWAYS	//flags: open for writing and create if file doesn't exist
					);
				
				char *text[100];
				uint8_t textLen=0;
				do
				{
					debug_msg("Enter a text (max 100 chars). Enter - accept. ESC - don't want to write any more\r\n");
					textLen = debug_input_string((char*)text,100);
					//add another text to the file
					uint16_t written=0;	//number of bytes written
					f_write(
						&file,	//pointer to file structure
						text,	//pointer to buffer that goes to the file
						textLen, //number of bytes to write
						&written); //pointer to a variable that stores number of bytes actually written
				}while(textLen);
				
				//close file
				f_close(&file);
				
				debug_msg("file is saved. Now you can read ");
				debug_txt(fileName);
				debug_txt(" with 'r' command");
				
				break;
			}//case 'W':
			case 'R':
			{
				debug_msg("Stream read from file. Enter file name: ");
				char fileName[FILE_NAME_LEN];
				debug_input_string(fileName, FILE_NAME_LEN);
				
				//create file structure
				FIL file;
				
				//open file for writing
				//create file if requested file doesn't exist
				FRESULT result = f_open(
					&file,						//a pointer to file
					(const char*)fileName,		//pointer to filename
					FA_READ	//flags: simply open for reading
					);
				
				//check result value
				if(result != FR_OK)	//on any error...
				{
					debug_msg("Could not open file ");
					debug_txt(fileName);
					//exit!
					break;
				}
				
				const uint8_t BYTES_PER_READ = 0x80;	//number of bytes that will be read at "one shot"
				char *buffer[BYTES_PER_READ];	//buffer to hold received data
				
				uint16_t bytesRead=0;	//number of bytes read from file
				do
				{
					//read another 0x80 bytes from file
					f_read(
						&file,	//pointer to file structure
						buffer,	//this buffer will be filled with data from file
						BYTES_PER_READ, //number of bytes to read
						&bytesRead //pointer to a variable that stores number of bytes actually written
					);
					
					//display buffer
					debug_dump( buffer , bytesRead );
					
					if(bytesRead == BYTES_PER_READ)
					{
						debug_msg("### Press Enter if you want to see more data. Esc to break");
					}
					else
					{
						debug_msg("### End of file. Press any key to exit.");
					}
					key = debug_waitkey();
				}while( (key != 27) && (bytesRead == BYTES_PER_READ) );	//exit loop on Esc OR end of file
				
				debug_msg("closing file...");
				
				//close file
				f_close(&file);
				
				debug_msg("File is closed now.");
				break;
			}//case 'R':
			case 'l':
			{
				debug_msg("Listing files in a directory. Enter directory path: ");
				char dirPath[20];
				debug_input_string(dirPath,20);
				fsListFiles(dirPath);
				break;
			}
			case 'r':
			{
				debug_msg("Reading file contents. Enter file name: ");
				char fileName[FILE_NAME_LEN];
				debug_input_string(fileName, FILE_NAME_LEN);
				debug_msg("enter starting byte (hex): ");
				uint16_t start = debug_input_hex(0);
				debug_msg("enter number of bytes to read (hex, max=200): ");
				uint16_t len =  debug_input_hex(1);
				if(len > 200)
				{
					debug_msg("truncating len to 200 bytes");
					len = 200;
				}
				uint8_t fileContents[200];
				
				len = fsRead(fileName,fileContents,start,len);
				
				if(len > 0)
				{
					debug_msg("File Read OK! File contents:");
					debug_dump(fileContents,len);
				}
				else
				{
					debug_msg("File read error!");
				}
				break;
			}//case 'r':
			case 'a':
			{
				debug_msg("append line to file. Enter file name: ");
				char fileName[FILE_NAME_LEN];
				debug_input_string(fileName, FILE_NAME_LEN);
				debug_msg("enter text to append (max 100 chars): ");
				char textToAppend[100];
				debug_input_string(textToAppend,100);
				if( fsAppend( fileName , (uint8_t*)textToAppend , strlen(textToAppend) ) == FR_OK )
				{
					debug_msg("text appended");
				}
				else
				{
					debug_msg("error");
				}
				break;
			}//case 'a':
			case 'o':
			{
				debug_msg("overwrite some bytes in file. Enter file name: ");
				char fileName[FILE_NAME_LEN];
				debug_input_string(fileName, FILE_NAME_LEN);
				debug_msg("enter starting byte:");
				uint32_t start = debug_input_hex(0);
				debug_msg("enter a new text (max 10 chars): ");
				char textThatOverwrites[10];
				debug_input_string(textThatOverwrites,10);
				if( fsReplace( fileName , (uint8_t*)textThatOverwrites , start , strlen(textThatOverwrites) ) == FR_OK )
				{
					debug_msg("text replaced!");
				}
				else
				{
					debug_msg("error");
				}
				break;
			}//case 'o':
			case 'd':
			{
				debug_msg("Remove file. Enter file name: ");
				char fileName[FILE_NAME_LEN];
				debug_input_string(fileName, FILE_NAME_LEN);
				fsRemove(fileName);
				debug_msg("done");
			}
			case 't':
			{
				debug_msg("truncate file. Enter file name: ");
				char fileName[FILE_NAME_LEN];
				debug_input_string(fileName, FILE_NAME_LEN);
				debug_msg("Enter number of bytes to remove (hex): ");
				uint32_t len = debug_input_hex(0);
				fsRemoveBytes(fileName,len);
				debug_msg("done");
				break;
			}
		}//switch(key)
	}//for(;;)
	return(0);
}


