#include "common.h"
#include "lcd_2x16.h"
#include "adc.h"


//Clock selection for Timer
#define TC_CLKS					0x7
#define TC_CLKS_MCK2			0x0
#define TC_CLKS_MCK8			0x1
#define TC_CLKS_MCK32			0x2
#define TC_CLKS_MCK128			0x3
#define TC_CLKS_MCK1024			0x4
#define SELECTED_TC_CLKS		TC_CLKS_MCK128
#define TC_PRESCALER			128

#define TIMER					AT91C_BASE_TC0
#define TIMER_ID				AT91C_ID_TC0


void adcIrqHandler(void);
void adcTimerInit(void);



void adcDmaStart(uint8_t channel, uint32_t samplingRate, void(*isrFunc)())
{
	// set ADC mode
    #if ADC_RESOLUTION_10_BIT==1
	
	const uint8_t ADC_PRESCAL = 4;
	const uint8_t ADC_STARTUP = 4;
	const uint8_t ADC_SHTIM = 4;
	
	ADC->ADC_MR =
		AT91C_ADC_TRGEN_EN |
        AT91C_ADC_TRGSEL_TIOA0 |
		AT91C_ADC_LOWRES_10_BIT |
		AT91C_ADC_SLEEP_NORMAL_MODE |
		((uint32_t)ADC_PRESCAL<<8) |		// prescale setting, fACD=5MHz=fmax
		((uint32_t)ADC_STARTUP<<16) |	// startup setting
		((uint32_t)ADC_SHTIM<<24);		// sample-and-hold time
		
    #else
	const uint8_t ADC_PRESCAL = 4;
	const uint8_t ADC_STARTUP = 4;
	const uint8_t ADC_SHTIM = 4;
	
	ADC->ADC_MR =
		AT91C_ADC_TRGEN_EN |
        AT91C_ADC_TRGSEL_TIOA0 |
		AT91C_ADC_LOWRES_8_BIT |
		AT91C_ADC_SLEEP_NORMAL_MODE |
		((uint32_t)ADC_PRESCAL<<8) |		// prescale setting, fACD=5MHz=fmax
		((uint32_t)ADC_STARTUP<<16) |	// startup setting
		((uint32_t)ADC_SHTIM<<24);		// sample-and-hold time
    #endif
	
	//reset DMA counters
	ADC->ADC_RCR = 0;
	ADC->ADC_RNCR = 0;
	
	//initialize Timer
    adcTimerInit();
	
	//set sampling rate before start
	adcSetSamplingRate(samplingRate);

	//mask channel (prevent forbidden values)
    channel = channel & 0x0F;
	
	//disable all registers - to be sure
    ADC->ADC_CHDR =	0xFF;
	
	//enable selected channel
    ADC->ADC_CHER =	1<<channel;
    
	//define ADC's IRQ priority
	const uint8_t ADC_INTERRUPT_PRIORITY = 2;
	
	//prepare AIC for ADC interrupts
	AT91F_AIC_ConfigureIt ( AT91C_BASE_AIC, ADC_ID, ADC_INTERRUPT_PRIORITY,AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL, isrFunc);
	
	//enable interrupt from ADC in AIC
	AT91F_AIC_EnableIt (AT91C_BASE_AIC, AT91C_ID_ADC);
	
	//enable ADC interrupt on end of buffer (ENDRX) or end of both buffers (RXBUFF)
	ADC->ADC_IER = AT91C_ADC_RXBUFF | AT91C_ADC_ENDRX;
	//interrupt comes here. Time to reinitialize buffers in ISR
	
	//start conversions
    ADC->ADC_CR = AT91C_ADC_START;
}


void adcDmaStop(void)
{
	//disable all interrupts from ADC
	ADC->ADC_IDR = 0xFFFFFFFF;
	
	//disable DMA RX transfer
	ADC->ADC_PTCR = AT91C_PDC_RXTDIS;
	
	//reset DMA counters
	ADC->ADC_RCR = 0;
	ADC->ADC_RNCR = 0;
}





void adcSamplingInit(uint8_t channel)
{
    #if ADC_RESOLUTION_10_BIT==1
	// set ADC mode
	ADC->ADC_MR =	AT91C_ADC_TRGEN_EN |
                                AT91C_ADC_TRGSEL_TIOA0 |
								AT91C_ADC_LOWRES_10_BIT |
								AT91C_ADC_SLEEP_NORMAL_MODE |
								(0x4<<8) |		// prescale setting, fACD=5MHz=fmax
								(0x04<<16) |	// startup setting
								(0x00<<24);		// sample-and-hold time
								
    #else
	ADC->ADC_MR =	AT91C_ADC_TRGEN_EN |
                                AT91C_ADC_TRGSEL_TIOA0 |
								AT91C_ADC_LOWRES_8_BIT |
								AT91C_ADC_SLEEP_NORMAL_MODE |
								(0x2<<8) |		// prescale setting, fACD=8MHz=fmax
								(0x08<<16) |	// startup setting
								(0x0<<24);		// sample-and-hold time
    #endif
	
	//initialize Timer
    adcTimerInit();

	//mask channel (prevent forbidden values)
    channel = channel & 0x0F;
	
	//disable all registers - to be sure
    ADC->ADC_CHDR =	0xFF;
	
	//enable selected channel
    ADC->ADC_CHER =	1<<channel;
    
	//define ADC's IRQ priority
	const uint8_t ADC_INTERRUPT_PRIORITY = 2;
	
	//prepare AIC for ADC interrupts
	AT91F_AIC_ConfigureIt ( AT91C_BASE_AIC, AT91C_ID_ADC, ADC_INTERRUPT_PRIORITY,1, adcIrqHandler);
	
	//enable interrupt from ADC in AIC
	AT91F_AIC_EnableIt (AT91C_BASE_AIC, AT91C_ID_ADC);
	
	//enable ADC interrupt on DRDY bit
	ADC->ADC_IER = AT91C_ADC_DRDY;
	
	//start conversions
    ADC->ADC_CR = AT91C_ADC_START;
}




void adcTimerInit(void)
{
    unsigned int dummy;
    const uint32_t INITIAL_SAMPLING_RATE = 500;	//[samples per second]

    // First, enable the clock of the TIMER
    AT91C_BASE_PMC->PMC_PCER = 1<<TIMER_ID;
	
    //disable timer's clock
	AT91C_BASE_TC0->TC_CCR = AT91C_TC_CLKDIS;
	
	//disable timer's interrupts
	AT91C_BASE_TC0->TC_IDR = 0xFFFFFFFF ;
	
	//read status register to clear its flags
	dummy = AT91C_BASE_TC0->TC_SR;
	dummy = dummy;
	
	//set capture mode register of the timer
	TIMER->TC_CMR = AT91C_TC_WAVE  //bit WAVE=1 => Waveform Mode
		|TC_CLKS_MCK128   //Timer clock freq. is MCK/128
		|AT91C_TC_ACPA_SET  //RA Compare Effect on TIOA: set (01)
		|AT91C_TC_ACPC_CLEAR  //RC Compare Effect on TIOA: clear (10)
		|AT91C_TC_WAVESEL_UP_AUTO;  //Waveform Selection: UP mode with automatic trigger on RC compare (10)
	
	adcSetSamplingRate(INITIAL_SAMPLING_RATE);
	
    //enable timer's clock
	TIMER->TC_CCR = AT91C_TC_CLKEN;
	
    //start timer
    TIMER->TC_CCR = AT91C_TC_SWTRG ;
}



uint8_t adcSetSamplingRate(uint32_t sps)
{
/*	#if ADC_RESOLUTION_10_BIT==1 //10 bit, Fsamax=384ksps
		if(sps>384000) return(0);
	#else //8 bit, Fsamax=533ksps
		if(sps>533000) return(0);
	#endif*/
	
	if(sps<SAMPLING_RATE_MIN) return(0);
	if(sps>SAMPLING_RATE_MAX) return(0);
	
	//compute value for timer full-count
	uint16_t TC_RCValue = MCK/TC_PRESCALER/sps;
	
	//prevent zero divider when using TC_RCValue / 2
	if(sps < 2) return(0);
	
	//write full-count value into RC
	TIMER->TC_RC = TC_RCValue;
	
	//write 0.5 of full value to RA
	TIMER->TC_RA = TC_RCValue/2;
	
	return(1);
}


void adcStopSampling(void)
{
    //enable timer's clock
	TIMER->TC_CCR = AT91C_TC_CLKDIS;
}


void adcStartSampling(void)
{
    //enable timer's clock
	TIMER->TC_CCR = AT91C_TC_CLKEN;
}






uint16_t adcGet(uint8_t channel)
{
	//define configuration fields in ADC_MR register
	const uint32_t SHTIM_MASK = 0x00 << 24;   //sample and hold time
	const uint32_t STARTUP_MASK = 0x04 << 16; //startup time
	const uint32_t PRESCAL_MASK = 0x4 << 8;   //prescaler setting
	
	//configure ADC
	ADC->ADC_MR = SHTIM_MASK | STARTUP_MASK | PRESCAL_MASK | AT91C_ADC_TRGEN_EN;
	
	//activate requested channel
    channel = channel & 0x07;
    ADC->ADC_CHER =	1<<channel;
	
	//start conversion!
	ADC->ADC_CR = AT91C_ADC_START;
	
	//wait for adc DRDY bit set
	while( ! (ADC->ADC_SR & AT91C_ADC_DRDY) );
	
	//read result from LCDR register
	uint16_t retval = ADC->ADC_LCDR;
	
	//disable all channels
	ADC->ADC_CHDR =	0xFF;
	
	return(retval);
}



void adcIrqHandler(void)
{
	LED_R_TOGGLE;
	uint16_t adcResult;
	adcResult = ADC->ADC_LCDR;
	lcdLocate(1,6);
	lcdValue(adcResult);
	lcdText("  ");
}




