#include "common.h"
#include "lcd_2x16.h"
#include "spi.h"
#include "sd_spi.h"
#include "fat/fs_tools.h"
#include "adc.h"

#define ADC_BUFFER_SIZE			4096
#define SAMPLING_RATE			16000


//wave file headers

//RIFF type chunk
typedef struct
{
	uint32_t chunkId;	//"RIFF" text
	uint32_t chunkDataSize;	//total file size - 8
	uint32_t riffType;	//"WAVE" text
}TRIFFHeader;

//Wave File chunks

//Format chunk (fmt)
typedef struct
{
	uint32_t chunkId;	//"fmt " text
	uint32_t chunkDataSize; //usually 0x16 (22 bytes)
	uint16_t compressionCode; //0x0001 for standard PCM format
	uint16_t numberOfChannels; //1 for mono signal, 2 for stereo and so on...
	uint32_t sampleRate; //how many samples per second (i.e. 44100 for audio CD)
	uint32_t averageBytesPerSecond; //how many bytes per second (ie. for stereo audio CD: 44100 16-bit samples gives 2*88200 bytes per second)
	uint16_t blockAlign; //number of bytes occupied by one sample
	uint16_t significantBitsPerSample; //number of bits per one sample in one channel
}TWaveFormatChunk;

//data chunk header, followed by data bytes
typedef struct
{
	uint32_t chunkId;	//"data" text
	uint32_t chunkDataSize; //number of data bytes following
}TWaveDataHeader;

//file object
volatile FIL wave;

//header structures
volatile TRIFFHeader			riffHeader;
volatile TWaveFormatChunk		waveFormatChunk;
volatile TWaveDataHeader		waveDataHeader;

//buffers for samples
volatile uint8_t adcBuffer[2][ADC_BUFFER_SIZE];
volatile uint8_t activeBuffer;

//prototype for ADC DMA ISR function
void adcDmaBufferSwap(void);


void wavCreateNewFile(const char *fileName)
{
	//prepare headers
	
	//wave main header
	riffHeader.chunkId = 0x46464952; //"RIFF" text, magic value
	riffHeader.chunkDataSize = 0;	//will be set after recording
	riffHeader.riffType = 0x45564157; //"WAVE" text, magic value
	
	//format chunk
	waveFormatChunk.chunkId = 0x20746D66; //"fmt " (space at end), magic value
	waveFormatChunk.chunkDataSize = 16; //size of data in this chunk
	waveFormatChunk.compressionCode = 0x0001; //0x0001 for standard PCM format
	waveFormatChunk.numberOfChannels = 0x0001; //0x0001 for mono signal
	waveFormatChunk.sampleRate = SAMPLING_RATE;	//number of samples per second
	waveFormatChunk.averageBytesPerSecond = SAMPLING_RATE; //number of bytes per second
	waveFormatChunk.blockAlign = 1; //one byte per sample
	waveFormatChunk.significantBitsPerSample = 8; //number of bits per one sample
	
	//data chunk header
	waveDataHeader.chunkId = 0x61746164;
	waveDataHeader.chunkDataSize = 0; //still zero, will be updated during recording
	
	//open file for writing
	//delete its contents if the file exists
	f_open(&wave,fileName,FA_WRITE|FA_CREATE_ALWAYS);
	
	//a variable that stores number of bytes written
	uint16_t written = 0;
	
	//write headers to file
	f_write(&wave,&riffHeader,sizeof(TRIFFHeader),&written);
	f_write(&wave,&waveFormatChunk,sizeof(TWaveFormatChunk),&written);
	f_write(&wave,&waveDataHeader,sizeof(TWaveDataHeader),&written);
}

inline void wavAddNewBuffer(void *buffer, uint16_t len)
{
	uint16_t written = 0;
	f_write(&wave,buffer,len,&written);
	waveDataHeader.chunkDataSize += written;
}


void wavFinishRecording(void)
{
	//a variable that stores number of bytes written
	uint16_t written = 0;
	
	//compute chunkDataSize from wave file main header
	//set it to file size - 8
	//file size is a sum of headers and data size
	riffHeader.chunkDataSize = 
		sizeof(TRIFFHeader) + sizeof(TWaveFormatChunk) + sizeof(TWaveDataHeader) //headers
		+ waveDataHeader.chunkDataSize	//plus number of data bytes
		- 8; //minus 8 bytes
	
	//set filepointer to start byte
	f_lseek(&wave,0);
	
	//write file headers again
	f_write(&wave,&riffHeader,sizeof(TRIFFHeader),&written);
	f_write(&wave,&waveFormatChunk,sizeof(TWaveFormatChunk),&written);
	f_write(&wave,&waveDataHeader,sizeof(TWaveDataHeader),&written);
	
	f_close(&wave);
}


void init(void)
{
	// enable peripheral clock for PIO controllers
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA);
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	debug_init_default();
	debug_msg("debug channel initialized");
	
	//initialize LEDs
	LED_R_INIT;
	LED_Y_INIT;
	LED_G_INIT;
	LED_B_INIT;
	
	spiInit();
	
	lcdInit();
	lcdLocate(1,0);
	lcdText("Voice Recorder");
	
	if(sdCardInit()==OK)
	{
		
		if(fsInit() == FR_OK)
		{
			lcdLocate(0,0);
			lcdText("SW1 to start");
		}
		else
		{
			lcdLocate(0,0);
			lcdText("No Filesystem!");
		}
		
	}
	else
	{
		lcdLocate(0,0);
		lcdText("Card Error!");
	}
}




int main( void )
{
	init();
	
	//define channel to sample
	enum
	{
		ADC_VR_CHANNEL = 4,
		ADC_MIC_CHANNEL = 7
	};
	//quick config
	const uint8_t ADC_CHANNEL_TO_SAMPLE = ADC_MIC_CHANNEL;
	uint8_t running = 0;
	

	for(;;)
	{
		if(SW3_PRESSED)
		{
			debug_msg("generating test file");
			wavCreateNewFile("test.wav");
			
			uint8_t aBuffer[512];
			uint16_t i;
			for(i=0;i<512;i++) aBuffer[i] = (uint8_t)i;
			
			wavAddNewBuffer(aBuffer,512);
			
			wavFinishRecording();
			
			debug_msg("test file generated");
			
		}
		
		if((SW1_PRESSED)&&(running==0))
		{
			//message for the user
			lcdLocate(0,0);
			lcdText("SW2 to stop ");
			
			//set flag
			running = 1;
			
			wavCreateNewFile("record.wav");
			
			//start DMA transfers from ADC
			adcDmaStart(ADC_CHANNEL_TO_SAMPLE, SAMPLING_RATE, adcDmaBufferSwap);
			
			LED_G_ON;
			
			//debounce pushbutton
			while(SW1_PRESSED);
			delay_ms(100);
		}
		
		if(SW2_PRESSED)
		{
			//message for the user
			lcdLocate(0,0);
			lcdText("SW1 to start");
			
			//clear flag
			running = 0;
			
			//stop DMA transfers from ADC
			adcDmaStop();
			
			wavFinishRecording();
			
			LED_G_OFF;
			
			//debounce pushbutton
			while(SW1_PRESSED);
			delay_ms(100);
		}
		
		LED_B_TOGGLE;
		delay_ms(50);
	}
	return(0);
	
}



//this function is called from adc.c module on ADC's RXBUFF and ENDRX flags
void adcDmaBufferSwap(void)
{

	//debug_msg("ISR");
	LED_G_OFF;
	
	//copy adc status to a variable
	static uint32_t status;
	status = ADC->ADC_SR;
	
	//both buffers are not initialized or full
	if(status & AT91C_ADC_RXBUFF)
	{
		debug_msg("DMA Init");
		activeBuffer = 0;
		ADC->ADC_RPR = (uint32_t) &adcBuffer[0][0];
		ADC->ADC_RNPR = (uint32_t) &adcBuffer[1][0];
		ADC->ADC_RCR = ADC_BUFFER_SIZE;
		ADC->ADC_RNCR = ADC_BUFFER_SIZE;
		ADC->ADC_PTCR = AT91C_PDC_RXTEN;
		return;
	}
	
	//buffer filling finished
	if(status & AT91C_ADC_ENDRX)
	{
		
		wavAddNewBuffer((void*)&adcBuffer[activeBuffer][0],ADC_BUFFER_SIZE);
		
		//prepare "next" buffer
		ADC->ADC_RNPR = (uint32_t) &adcBuffer[activeBuffer][0];
		ADC->ADC_RNCR = ADC_BUFFER_SIZE;
		
		if(activeBuffer)
		{
			//debug_msg("RXBUFF 0");
			activeBuffer = 0;
		}
		else
		{
			//debug_msg("RXBUFF 1");
			activeBuffer = 1;
		}
	}
	
	LED_G_ON;
}


