#include "common.h"
#include "lcd_2x16.h"
#include "spi.h"
#include "sd_spi.h"
#include "fat/fs_tools.h"
#include "adc.h"
#include "pwm_dac.h"

#define DAC_BUFFER_SIZE			4096
#define SAMPLING_RATE			16000


//wave file headers

//RIFF type chunk
typedef struct
{
	uint32_t chunkId;	//"RIFF" text
	uint32_t chunkDataSize;	//total .wav file size - 8
	uint32_t riffType;	//"WAVE" text
}TRIFFHeader;

//Wave File chunks

//Format chunk (fmt)
typedef struct
{
	uint32_t chunkId;	//"fmt " text
	uint32_t chunkDataSize; //usually 0x16 (22 bytes)
	uint16_t compressionCode; //0x0001 for standard PCM format
	uint16_t numberOfChannels; //1 for mono signal, 2 for stereo and so on...
	uint32_t sampleRate; //how many samples per second (i.e. 44100 for audio CD)
	uint32_t averageBytesPerSecond; //how many bytes per second (ie. for stereo audio CD: 44100 16-bit samples gives 2*88200 bytes per second)
	uint16_t blockAlign; //number of bytes occupied by one sample
	uint16_t significantBitsPerSample; //number of bits per one sample in one channel
}TWaveFormatChunk;

//data chunk header, followed by data bytes
typedef struct
{
	uint32_t chunkId;	//"data" text
	uint32_t chunkDataSize; //number of data bytes following
}TWaveDataHeader;

//file object
FIL wave;

//header structures
TRIFFHeader			riffHeader;
TWaveFormatChunk	waveFormatChunk;
TWaveDataHeader		waveDataHeader;

//file name buffer
const char waveFile[] = "record.wav";

//buffers for samples
static volatile int8_t dacBuffer[2][DAC_BUFFER_SIZE];
static volatile uint8_t activeBuffer = 0;
static volatile uint8_t bufferToProcess = 1;
static volatile uint8_t dataRequest;
static volatile uint16_t dacSample;
static volatile uint8_t dacActive;

void pwmdacIsr(void);




void init(void)
{
	// enable peripheral clock for PIO controllers
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA);
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	debug_init_default();
	debug_msg("debug channel initialized");
	
	//initialize LEDs
	LED_R_INIT;
	LED_Y_INIT;
	LED_G_INIT;
	LED_B_INIT;
	
	spiInit();
	
	lcdInit();
	lcdLocate(1,0);
	lcdText("Wave Player");
	
	if(sdCardInit()==OK)
	{
		
		if(fsInit() == FR_OK)
		{
			lcdLocate(0,0);
			lcdText("SW1 to start");
		}
		else
		{
			lcdLocate(0,0);
			lcdText("No Filesystem!");
		}
		
	}
	else
	{
		lcdLocate(0,0);
		lcdText("Card Error!");
	}
}




int main( void )
{
	init();
	
	pwmdacInit(SAMPLING_RATE,pwmdacIsr);
	
	for(;;)
	{
		if((SW1_PRESSED)&&(dacActive==0))
		{
			if(f_open(&wave,"record.wav",FA_READ) != FR_OK)
			{
				lcdLocate(0,0);
				lcdText("no file!    ");
				continue;
			}
			else
			{
				lcdLocate(0,0);
				lcdText("playing...  ");
			}
			
			uint16_t read=0;
			//read headers, data in them may be useful
			f_read(&wave,&riffHeader,sizeof(TRIFFHeader),&read);
			f_read(&wave,&waveFormatChunk,sizeof(TWaveFormatChunk),&read);
			f_read(&wave,&waveDataHeader,sizeof(TWaveDataHeader),&read);
			//debounce pushbutton
			while(SW1_PRESSED);
			delay_ms(100);
			
			activeBuffer = 0;
			bufferToProcess = 1;
			
			//set flag
			dacActive = 1;
		}
		
		
		if(dataRequest)
		{
			//debug_msg("dreq!");
			dataRequest = 0;
			uint16_t dataRead = 0;
			f_read(&wave,(void*)&dacBuffer[bufferToProcess][0],DAC_BUFFER_SIZE,&dataRead);
			if(dataRead==0)
			{
				debug_msg("end of file");
				dacActive = 0;
				
				//close file
				f_close(&wave);
				
				//message for the user
				lcdLocate(0,0);
				lcdText("SW1 to start");
			}
		}
		
	}
	return(0);
	
}



void pwmdacIsr(void)
{
	uint32_t status = pwmdacGetTimerStatus();
	status = status;
	
	LED_G_OFF;
	if(dacActive==0) return;
	LED_G_ON;
	
	if(dacSample < DAC_BUFFER_SIZE)
	{
		//set dac
		pwmdacSet(dacBuffer[activeBuffer][dacSample]);
		dacSample++;
	}
	else	//swap buffers and get data
	{
		dacSample = 0;
		if(activeBuffer)
		{
			activeBuffer = 0;
			bufferToProcess = 1;
		}
		else
		{
			activeBuffer = 1;
			bufferToProcess = 0;
		}
		pwmdacSet(dacBuffer[activeBuffer][dacSample]);
		dataRequest = 1;	//set flag to inform main loop
	}
	
}

