#include "common.h"
#include "dbgu.h"
#include "lcd_2x16.h"
#include "one_wire.h"
#include "ds18b20.h"
#include "spi.h"
#include "sd_spi.h"
#include "fat/fs_tools.h"
#include "mass_storage/msc_ui.h"


void init(void)
{
	// enable peripheral clock for PIO controllers
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOA);
	AT91C_BASE_PMC->PMC_PCER = (1<<AT91C_ID_PIOB);
	
	// enable reset-key on board 
	AT91C_BASE_RSTC->RSTC_RMR = (0xA5000000 | AT91C_RSTC_URSTEN);
	
	//open debug channel
	debug_init_default();
	
	//initialize LEDs
	LED_R_INIT;
	LED_Y_INIT;
	LED_G_INIT;
	LED_B_INIT;
	
	//init 1-Wire
	owInit();
	
	//initialize SPI interface
	spiInit();
	
	//initialize 2x16 LCD
	lcdInit();
	
	//initialize SD Card
	if(sdCardInit() == OK)
	{
		debug_msg("SD card initialized OK");
	}
	else
	{
		debug_msg("SD card initialization FAILED!");
		lcdLocate(0,0);
		lcdText("Card Init Fail");
		while(1);
	}
	
	//mount filesystem
	if(fsInit()!=FR_OK)
	{
		lcdLocate(0,0);
		lcdText("No Filesystem");
		while(1);
	}
	
	//initialize Mass Storage Device stuff
	mscInit(0);
	
}

#define FILE_NAME_LEN		(8+1+3+1) //8 chars name, 1 char dot, 3 chars extension, 1 char NULL

int main( void )
{
	init();
	
	lcdCls();
	lcdLocate(0,0);
	lcdText("Temp. Logger");
	lcdLocate(1,0);
	lcdText("  Logger mode   ");
	delay_ms(1000);
	lcdLocate(0,0);
	lcdText("                ");
	
	uint8_t rom[8];
	
	debug_msg("temperature logger");
	
	enum {
		STATE_LOG = 0,
		STATE_USB = 1
	};
	
	uint8_t previousState = STATE_LOG;
	
	for(;;)
	{
		
		if(mscPlugInserted())
		{
			//enter this "if" only once after plug insertion
			if(previousState == STATE_LOG)
			{
				lcdLocate(0,0);
				lcdText("                ");
				lcdLocate(1,0);
				lcdText(" Host connected ");
				//initialize USB transmission
				mscStart();
			}
			
			//call it in a loop
			mscUIEnumerate();
			
			previousState = STATE_USB;
		}
		else
		{
			//enter this "if" only once after USB plug removal
			if(previousState == STATE_USB)
			{
				lcdLocate(0,0);
				lcdText("                ");
				lcdLocate(1,0);
				lcdText("  Logger mode   ");
				
				//terminate USB transmission
				mscStop();
				
				//reinitialize fileSystem
				if(fsInit()!=FR_OK)
				{
					lcdLocate(1,0);
					lcdText("Filesystem error");
					continue;
				}
			}
			
			previousState = STATE_LOG;
			
			//rest of the Logger's code
			
			if(owReadRom(rom)!=OK)
			{
				lcdLocate(0,0);
				lcdText("Connect DS18B20!");
				delay_ms(100);
				continue;
			}
			
			
			//rom is ok, read temperature
			ds18b20ConvertT();	//start temperature measurement
			delay_ms(1000);
			int16_t temperature = ds18b20ReadTempNoRom(); //get temperature from sensor
			
			//convert DS18B20 temperature register contents into readable text
			
			int8_t base = temperature >> 4;	//extract integer part
			
			uint8_t decimal = temperature & 0x0F;	//extract decimal part from bits 3...0
			decimal = (decimal * 10) >> 4;			//muliply decimal part by 10/16
			
			char logText[9] = "T=";
			if( temperature < 0 )
			{
				base = - base - 1;
				decimal = 10 - decimal;
				logText[2] = '-';
			}
			else
			{
				logText[2] = '+';
			}
			
			logText[3] = '0' + base / 100;
			base = base % 100;
			logText[4] = '0' + base / 10;
			base = base % 10;
			logText[5] = '0' + base;
			logText[6] = '.';
			logText[7] = '0' + decimal;
			logText[8] = 0;	//end of string
			
			lcdLocate(0,0); //set LCD cursor to upper line
			lcdText(logText); lcdText("  ");
			//debug_msg("text to write: "); debug_txt(logText);
			fsWriteLine("TEMPLOG.TXT",logText);
			
		}
	}
	return(0);
}


