#include "Board.h"
#include "common.h"
#include "dbgu.h"

#if defined AT91SAM7X128 || defined AT91SAM7X256 || defined AT91SAM7XC256 || defined AT91SAM7XC128
	#define DBGU_PIO				AT91C_BASE_PIOA
	#define DBGU_RX_MASK			AT91C_PA27_DRXD
	#define DBGU_TX_MASK			AT91C_PA28_DTXD
#endif

#if defined AT91SAM7S64 || defined AT91SAM7S128 || defined AT91SAM7S256
	#define DBGU_PIO				AT91C_BASE_PIOA
	#define DBGU_RX_MASK			AT91C_PA9_DRXD
	#define DBGU_TX_MASK			AT91C_PA10_DTXD
#endif

#define DBGU_BAUDRATE			115200
AT91PS_DBGU 					pDbgu;

//private tools
static uint8_t debug_dec2hex(uint8_t val);
static uint8_t debug_char_is_hex(uint8_t chr);
static uint8_t debug_hex2dec(uint8_t val);
static uint32_t *debug_base_address;


void debug_init_default(void)
{
	debug_init(AT91C_BASE_DBGU, DBGU_PIO, DBGU_RX_MASK | DBGU_TX_MASK, 0);
}

//initializes pDbgu port for simple UART operation on 115200 baud (platform dependent)
//requires:
//paramDbgu: base address of DBGU,
//paramPio: base address of PIO pins used for RX and TX signals
//pinMask: mask for RX and TX pins
//periph: zero for peripheral A, non-zero for peripheral B
void debug_init(AT91PS_DBGU paramDbgu, AT91PS_PIO paramPio, uint32_t pinMask, uint8_t periph)
{
	//copy base address of USART to the global variable pDbgu
	pDbgu = paramDbgu;
	//disable PIO on pDbgu pins
	paramPio->PIO_PDR = pinMask;
	if(periph==0)
	{
		//select peripheral A on pDbgu pins
		paramPio->PIO_ASR = pinMask;
	}
	else
	{
		//select peripheral B on pDbgu pins
		paramPio->PIO_BSR = pinMask;
	}
	//reset and disable transmitter and receiver
	pDbgu->DBGU_CR = AT91C_US_RSTRX | AT91C_US_RSTTX | AT91C_US_RXDIS | AT91C_US_TXDIS;
	//set mode
	pDbgu->DBGU_MR = AT91C_US_PAR_NONE;
	//set baud rate
	pDbgu->DBGU_BRGR = AT91F_US_Baudrate(MCK,DBGU_BAUDRATE);
	//enable transmitter and receiver
	pDbgu->DBGU_CR = AT91C_US_RXEN | AT91C_US_TXEN;
}


//send chr via pDbgu UART (platform dependent)
void debug_chr(char chr)
{
	while ( !(pDbgu->DBGU_CSR & AT91C_US_TXRDY) );	//Wait for empty TX buffer
	pDbgu->DBGU_THR = chr;								//Transmit char
}


//returns ascii value of last char received
//returns 0 if no char was received since last debug_inkey call
//(platform dependent)
char debug_inkey(void)
{
	if( pDbgu->DBGU_CSR & AT91C_US_RXRDY )
	{
		return( pDbgu->DBGU_RHR );
	}
	else
	{
		return(0);
	}
}


//platform independent funcs


//prints text starting at str
//adds new line first
void debug_msg(char *str)
{
	debug_chr('\r');
	debug_chr('\n');
	debug_txt(str);
}

//prints text starting at str
void debug_txt(char *str)
{
	while(*str) debug_chr(*str++);
}


//prints text starting at str
//prints exactly len chars
void debug_txt_limit(char *str, uint8_t len)
{
	while(len)
	{
		debug_ascii(*str);
		str++;
		len--;
	}
}


//halts program/task execution until char is received
char debug_waitkey(void)
{
	char chr;
	do
	{
		chr = debug_inkey();
	}while(chr==0);
	return(chr);
}


//sends char b over pDbgu UART. Replaces values that can change cursor pos. on terminal
void debug_ascii(uint8_t b)
{
	switch(b)
	{
		case 0:
		{
			debug_chr('.');	//replace 0 with dot
			break;
		}
		case 8:
		case 9:
		case 10:
		case 13:
		{
			//avoid other chars that modify terminal cursor
			//replace them with space
			debug_chr(' ');
			break;
		}
		default:
		{
			debug_chr(b);
		}
	}//switch(chr)
}

//displays hex number from val in format FF
void debug8_t(uint8_t val)
{
	debug_chr(debug_dec2hex(val>>4));
	debug_chr(debug_dec2hex(val&0x0F));
}


//displays hex number from val in format FFFF
void debug16_t (uint16_t val)
{
	debug8_t(val>>(1*8));
	debug8_t(val>>(0*8));
}

//displays hex number from val in format FFFFFFFF
void debug32_t (uint32_t val)
{
	debug8_t(val>>(3*8));
	debug8_t(val>>(2*8));
	debug8_t(val>>(1*8));
	debug8_t(val>>(0*8));
}

//displays unsigned decimal number
void debug_word(uint16_t vtp)
{
	uint16_t temp1=0;
	
	debug_chr(vtp/10000+0x30);
	temp1 = (vtp%10000);
	debug_chr(temp1/1000+0x30);
	temp1 = temp1%1000;
	debug_chr(temp1/100+0x30);
	temp1 = temp1%100;
	debug_chr(temp1/10+0x30);
	temp1 = temp1%10;
	debug_chr(temp1+0x30);
}

//displays signed decimal number
void debug_int(int16_t vtp)
{
	uint16_t temp1=0;
	
	if((vtp&0x8000)==0x8000)
	{
		debug_chr('-');
		vtp = (~vtp) + 1;
	}
	else
	{
		debug_chr('+');
	}
	debug_chr(vtp/10000+0x30);
	temp1 = (vtp%10000);
	debug_chr(temp1/1000+0x30);
	temp1 = temp1%1000;
	debug_chr(temp1/100+0x30);
	temp1 = temp1%100;
	debug_chr(temp1/10+0x30);
	temp1 = temp1%10;
	debug_chr(temp1+0x30);
}

//converts decimal digit from into ascii hex digit
//returns ascii char of hex digit
static uint8_t debug_dec2hex(uint8_t val)
{
	if(val<10) val+='0';
	else val=val-10+'A';
	return(val);
}

//converts hex digit in ASCII format (val) into integer (return)
static uint8_t debug_hex2dec(uint8_t val)
{
	if((val>='a')&&(val<='f'))
		return(val+10-'a');
	if((val>='A')&&(val<='F'))
		return(val+10-'A');
	return(val-'0');
}

//prints text from *s and value
void debug_value( const char* s, uint32_t value)
{
    debug_txt("\n\r");
    debug_txt((char*)s);
    debug_txt("0x");
    if(value>0x00FFFFFF) debug8_t(value>>(3*8));
    if(value>0x0000FFFF) debug8_t(value>>(2*8));
    if(value>0x000000FF) debug8_t(value>>(1*8));
    debug8_t(value>>(0*8));
}

//prints array contents byte-by-byte in ASCII hex format
//provide starting address in *array and number of items to display in size
void debug_array( void *array, uint16_t size )
{
	uint8_t *s = array;
    debug_txt("\n\r");
	debug_txt("Length="); debug16_t(size); debug_txt(", Contents:\n\r");
	do
	{
		debug_chr(' '); debug8_t(*s++);
		size--;
	}while(size);
    debug_txt(" end");
}


//input:
//	BufSize - for buffer limitation
//	pBuf - pointer to string buffer
//returns:
//	length of obtained text (0 means no text)
uint8_t debug_input_string(char * pBuf, uint8_t BufSize)
{
	enum
	{
		BACKSPACE = 8,
		RET = 13,
		ESC = 27
	};
//	const uint8_t REFRESH_DELAY = 5;
	uint8_t Cursor=0;
	uint8_t rxedChar;
	pBuf[0]=0;
	
	for(;;)
	{
		rxedChar = debug_inkey();
		if(rxedChar)
		{
			switch(rxedChar)
			{
				case BACKSPACE:
				{
					if(Cursor)
					{
						Cursor--;
						pBuf[Cursor] = 0;
						debug_chr(rxedChar);
					}
					break;
				}
				case RET:
				{
					pBuf[Cursor]=0; //mark end of string
					return(Cursor);
				}
				case ESC:
				{
					return(0); //return zero length
				}
				default:
				{
					pBuf[Cursor]=rxedChar;
					if(Cursor<(BufSize-1))
					{
						Cursor++;
						debug_chr(rxedChar);
					}
				}
			}//switch(rxedChar)
		}//if(rxedChar)
		//vTaskDelay(REFRESH_DELAY);
	}//for(;;)
	
	return(0);
}



//returns 1 if chr is ascii HEX number else returns zero
static uint8_t debug_char_is_hex(uint8_t chr)
{
	if( ((chr>=0x30)&&(chr<=0x39)) || ( (chr>='a')&&(chr<='f') ) || ( (chr>='A')&&(chr<='F') ) ) return(1); else return(0);
}


//provide pointer to hext number in ASCII format *pBuf
//returns value of ascii number in buffer
uint32_t debug_ascii_to_int_hex(char *pBuf)
{
	uint32_t output=0;
	uint8_t pos=strlen(pBuf);
	for(uint8_t i=0;i<8;i++)
	{
		if(pos) pos--; else return(output);
		output+=((debug_hex2dec(pBuf[pos]))<<i*4);
	}
	return(output);
}


//returns input value on fail (no change) and new value on OK
uint32_t debug_input_hex(uint32_t input)
{
	const uint8_t BUFSIZE = 10;
	uint32_t Value=0;
	uint8_t SizeObtained=0;
	
	//allocate memory
	//uint8_t *pBuf;
	//pBuf = (uint8_t *) pvPortMalloc( BUFSIZE*sizeof(uint8_t) );
	uint8_t pBuf[BUFSIZE];
	for(uint8_t i=0;i<BUFSIZE;i++) pBuf[i]=0;
	
	//get string from terminal
	SizeObtained = debug_input_string((char*)pBuf, BUFSIZE);
	
	//check if length > 0
	if(SizeObtained==0) return(input);
	
	//scan buffer if contains digits only
	for(uint8_t i=0;i<SizeObtained;i++)
	{
		if(!debug_char_is_hex(pBuf[i])) return(input);
	}
	
	//convert number string to integer
	Value = debug_ascii_to_int_hex((char*)pBuf);
	
	//free allocated memory region
	//vPortFree(pBuf);
	
	//confirm a new value
	//debug_value("using value = ",Value);
	
	//return a new value
	return(Value);
}

void debug_dump(void *address, uint16_t len)
{
	uint8_t *buf = address;
	const uint16_t bytesInLine = 16;
	const uint16_t spaceBetweenDumpAndASCII = 4;
	uint16_t i, counter=len;
	
	//header
	debug_msg("Memory dump, starting address = "); debug32_t((uint32_t)buf); debug_txt(", length = "); debug16_t(len);
	debug_msg("Address  Offs:   Data");
	//contents
	while(1)
	{
		//insert last line (may be shorter than full line)
		if(counter < bytesInLine)
		{
			debug_txt("\r\n"); debug32_t((uint32_t)buf); debug_txt(" "); debug16_t(len-counter); debug_txt(":   ");
			
			//contents in hex
			for(i=0;i<bytesInLine;i++)
			{
				if(i<counter)
				{
					debug8_t(buf[i]);
					debug_chr(' ');
				}
				else
				{
					debug_txt("   ");
				}
				if(i%8==7) debug_chr(' ');
			}
			
			//space
			for(i=0;i<spaceBetweenDumpAndASCII;i++)
			{
				debug_chr(' ');
			}
			
			//contents in ASCII
			for(i=0;i<bytesInLine;i++)
			{
				if(i<counter)
				{
					debug_ascii(buf[i]);
				}
				else
				{
					debug_chr(' ');
				}
			}
			
			break;
		}
		
		debug_txt("\r\n"); debug32_t((uint32_t)buf); debug_txt(" "); debug16_t(len-counter); debug_txt(":   ");
		
		for(i=0;i<bytesInLine;i++)
		{
			debug8_t(buf[i]); debug_chr(' ');
			if(i%8==7) debug_chr(' ');
		}
		
		//space
		for(i=0;i<spaceBetweenDumpAndASCII;i++)
		{
			debug_chr(' ');
		}
		
		//contents in ASCII
		for(i=0;i<bytesInLine;i++)
		{
			debug_ascii(buf[i]);
		}
		
		buf += bytesInLine;
		if(counter >= bytesInLine)
		{
			counter -= bytesInLine;
		}
		
		if(counter == 0) break;
		
	}	//while(counter)
	//footer
	debug_msg("End of dump");
}


void debug_write_memory(void)
{
	debug_txt("reading memory, address: ");
	uint32_t *location;
	location = (uint32_t*) debug_input_hex(0xFFFFFFFF);
	
	if(location == (uint32_t*) 0xFFFFFFFF)
	{
		debug_txt("aborted");
		return;
	}
	
	debug_txt(", value: ");
	uint32_t value;
	value = debug_input_hex(0);
	
	*location = value;
}

void debug_read_memory(void)
{
	debug_txt("reading memory, address: ");
	uint32_t *location;
	location = (uint32_t*) debug_input_hex(0xFFFFFFFF);
	debug_txt(", value: ");
	debug32_t(*location);
}

void debug_write_reg(void)
{
	debug_txt("writing register, base: ");
		debug32_t((uint32_t)debug_base_address);
	debug_txt(", offset: ");
	uint32_t *offset;
	offset = (uint32_t*) debug_input_hex(0xFFFFFFFF);
	if(offset == (uint32_t*) 0xFFFFFFFF)
	{
		debug_txt(" aborted");
	}
	debug_txt(", value: ");
	uint32_t value;
	value = debug_input_hex(0);
	uint32_t *destination;
	destination = (uint32_t*) ( (uint32_t)debug_base_address + (uint32_t)offset );
	*destination = value;
}

void debug_read_reg(void)
{
	debug_txt("reading register, base: ");
		debug32_t((uint32_t)debug_base_address);
	debug_txt(", offset: ");
	uint32_t *offset;
	offset = (uint32_t*) debug_input_hex(0);
	uint32_t *destination;
	destination = (uint32_t*) ( (uint32_t)debug_base_address + (uint32_t)offset );
	debug_txt(", value: ");
	debug32_t(*destination);
}

void debug_set_base(void)
{
	debug_txt("new base address: ");
	debug_base_address = (uint32_t*) debug_input_hex((uint32_t)debug_base_address);
}

void debug_menu(void)
{
	uint8_t key;
	debug_msg("### ");
	do
	{
		key = debug_inkey();
		if(key)
		{
			switch(key)
			{
				case 'B':
				{
					debug_set_base();
					break;
				}
				case 'd':
				{
					debug_txt("dump 256 bytes from base address");
					debug_dump(debug_base_address,256);
					break;
				}
				case 'D':
				{
					debug_txt("dump data. Start address: ");
					uint32_t *startAddress;
					startAddress = (uint32_t*) debug_input_hex(0);
					debug_txt(" length: ");
					uint16_t len;
					len = debug_input_hex(1);
					debug_dump(startAddress,len);
					break;
				}
				case 'W':
				{
					debug_write_memory();
					break;
				}
				case 'w':
				{
					debug_write_reg();
					break;
				}
				case 'R':
				{
					debug_read_memory();
					break;
				}
				case 'r':
				{
					debug_read_reg();
					break;
				}
				case '?':	//display help
				{
					debug_txt("debug menu help");
					debug_msg(" B - set global base address");
					debug_msg(" W - write memory location");
					debug_msg(" R - read memory location");
					debug_msg(" w - write memory location (relative)");
					debug_msg(" r - read memory location (relative)");
					debug_msg(" D - parametric dump");
					debug_msg(" d - dump 256 bytes from base address");
					break;
				}
			}//switch(key)
			debug_msg("### ");
		}//if(key)
	}while( key != 27 );	//27 is ESC
}


