#include "common.h"
#include "lcd_2x16.h"
#include "usart_dma_irq.h"

#define USART_BASE			AT91C_BASE_US0
#define USART_ID			AT91C_ID_US0
#define USART_BAUDRATE		57600
const uint16_t USART_TIMEOUT = 57600;	//timeout (in bits)
#define AIC_BASE			AT91C_BASE_AIC	//AIC base address
#define USART_PIO			AT91C_BASE_PIOA	//PIO controller that muxes USART pins
#define PIO_AB_SEL_REG		PIO_ASR			//select PIO_ASR or PIO_BSR
#define USART_RX_MASK		AT91C_PA0_RXD0	//PIO pin multiplexed with RXD line
#define USART_TX_MASK		AT91C_PA1_TXD0	//PIO pin multiplexed with TXD line
#define USART_RTS_MASK		AT91C_PA3_RTS0	//PIO pin multiplexed with RTS line
#define USART_CTS_MASK		AT91C_PA4_CTS0	//PIO pin multiplexed with CTS line

static uint16_t maxBufferLength;
static uint8_t *buffer;
static volatile uint16_t receivedDataCounter;

//prototype for USART Interrupt Service Routine
void usartIsr(void);


//send "txBuffer" via usart
//"len" is number of bytes to send
void usartDmaIrqTransmit(uint8_t *txBuffer, uint16_t len)
{

	//print a message
	debug_msg("DMA Tx transfer...");
	lcdLocate(0,0); lcdText("DMA TX...       ");
	
	//write starting memory location
	USART_BASE->US_TPR = (uint32_t) txBuffer;
	
	//set number of bytes to transmit
	USART_BASE->US_TCR = len;
	
	//start the DMA TX transfer!
	USART_BASE->US_PTCR = AT91C_PDC_TXTEN;
	
	//enable DMA TX interrupt
	USART_BASE->US_IER = AT91C_US_TXBUFE;
}


//returns 0 if no data is waiting
//or
//copies received data into "rxBuffer" and returns number of bytes received
uint16_t usartDmaIrqReceive(uint8_t *rxBuffer)
{
	//disable interrupts from USART
	AT91F_AIC_DisableIt ( AT91C_BASE_AIC , USART_ID );
	
	//make a copy of receivedDataCounter flag
	uint16_t retval = receivedDataCounter;
	
	if(receivedDataCounter)
	{
		uint16_t i;
		//copy data to another buffer
		for(i=0;i<receivedDataCounter;i++)
		{
			rxBuffer[i] = buffer[i];
		}
		
		//reset counter flag
		receivedDataCounter = 0;
		
		//initialize rx buffer pointer
		USART_BASE->US_RPR = (uint32_t) buffer;
		
		//restart DMA rx transfer!
		USART_BASE->US_PTCR = AT91C_PDC_RXTEN;
		
		//reinitialize DMA rx counter
		USART_BASE->US_RCR = maxBufferLength;
		
		//reenable DMA rx complete interrupt
		USART_BASE->US_IER = AT91C_US_RXBUFF;
	}
	
	//reenable interrupts from USART
	AT91F_AIC_EnableIt ( AT91C_BASE_AIC , USART_ID );
	return(retval);
	
}


//initialization func
void usartDmaIrqInit(uint8_t *paramBuffer, uint16_t paramMaxLen)
{
	//copy parameters to global variables
	maxBufferLength = paramMaxLen;
	buffer = paramBuffer;
	
	//initialize rx data counter
	receivedDataCounter = 0;
	
	//enable USART clock
	AT91C_BASE_PMC->PMC_PCER = (1<<USART_ID);
	
	//disable PIO control on USART RX/TX pins
	USART_PIO->PIO_PDR = USART_RX_MASK | USART_TX_MASK | USART_RTS_MASK | USART_CTS_MASK;
	
	//select peripheral A or B register for USART pins
	USART_PIO->PIO_AB_SEL_REG = USART_RX_MASK | USART_TX_MASK | USART_RTS_MASK | USART_CTS_MASK;
	
	//reset and disable transmitter and receiver
	USART_BASE->US_CR = AT91C_US_RSTRX | AT91C_US_RSTTX |
		AT91C_US_RXDIS | AT91C_US_TXDIS | AT91C_US_RTSDIS;
	
	//set baud rate
	USART_BASE->US_BRGR = AT91F_US_Baudrate(MCK,USART_BAUDRATE);
	//USART_BASE->US_BRGR = MCK / 16 / USART_BAUDRATE;
	
	//set mode of operation
	USART_BASE->US_MR = AT91C_US_USMODE_HWHSH | AT91C_US_PAR_NONE | AT91C_US_CHRL_8_BITS;
	
	//set default timeout value
	USART_BASE->US_RTOR = USART_TIMEOUT;
	
	//reset timeout counter
	USART_BASE->US_CR = AT91C_US_STTTO;
	
	//initialize DMA TX counters
	USART_BASE->US_TCR = 1; //any nonzero value
	USART_BASE->US_TNCR = 0;
	
	//initialize DMA RX counters
	USART_BASE->US_RCR = maxBufferLength;
	USART_BASE->US_RNCR = 0;
	
	//initialize rx buffer pointers
	USART_BASE->US_RPR = (uint32_t) buffer;
	
	//configure AIC to work with usart
	const uint8_t USART_IRQ_PRIORITY = 5;
	AT91F_AIC_ConfigureIt ( AIC_BASE , USART_ID , USART_IRQ_PRIORITY , AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL , usartIsr );
	//edge sensitive mode is not recommended
	
	//enable USART interrupt source in AIC
	AT91F_AIC_EnableIt ( AIC_BASE , USART_ID );
	
	//enable interrupt sources form USART (select bits that will generate an interrupt)
	USART_BASE->US_IER = AT91C_US_RXBUFF | AT91C_US_TIMEOUT;
	
	//enable transmitter and receiver
	USART_BASE->US_CR = AT91C_US_RXEN | AT91C_US_TXEN | AT91C_US_RTSEN;
	
	//start dma rx
	USART_BASE->US_PTCR = AT91C_PDC_RXTEN;
}


//usart interrupt service routine
void usartIsr(void)
{
	//copy USART status reg to a variable
	uint32_t status;
	status = USART_BASE->US_CSR;
	
	//buffer tx done
	if(status & AT91C_US_TXBUFE)
	{
		debug_msg("ISR: TXBUFE!");
		lcdLocate(0,0); lcdText("ISR: TXBUFE!    ");
		
		//disable DMA TX interrupt
		USART_BASE->US_IDR = AT91C_US_TXBUFE;
		
		//disable TX DMA transfer
		USART_BASE->US_PTCR = AT91C_PDC_TXTDIS;
		
		//write any nonzero value to TCR not to get TXBUFE flag every time
		USART_BASE->US_TCR = maxBufferLength;
	}
	
	//rx timeout detected
	if( (status & AT91C_US_TIMEOUT) )
	{
		USART_BASE->US_PTCR = AT91C_PDC_RXTDIS;
		
		//reset timeout counter
		USART_BASE->US_CR = AT91C_US_STTTO;
		
		//compute how many bytes are received
		receivedDataCounter = maxBufferLength - USART_BASE->US_RCR;
		
		//disable interrupt from rx DMA complete and from timeout
		USART_BASE->US_IDR = AT91C_US_RXBUFF;
		
		debug_msg("ISR: TIMEOUT!");
		lcdLocate(0,0); lcdText("ISR: TIMEOUT!   ");
	}
	
	if( status & AT91C_US_RXBUFF )
	{
		USART_BASE->US_PTCR = AT91C_PDC_RXTDIS;
		
		//reset timeout counter
		USART_BASE->US_CR = AT91C_US_STTTO;
		
		//compute how many bytes are received
		receivedDataCounter = maxBufferLength;
		
		//disable interrupt from rx DMA complete and from timeout
		USART_BASE->US_IDR = AT91C_US_RXBUFF;
		
		debug_msg("ISR: RXBUFF!");
		lcdLocate(0,0); lcdText("ISR: RXBUFF!    ");
	}
}


