#include "common.h"
#include "lcd_2x16.h"
#include "usart_irq.h"

#define USART_BASE			AT91C_BASE_US0
#define USART_ID			AT91C_ID_US0
#define USART_BAUDRATE		115200
#define AIC_BASE			AT91C_BASE_AIC	//AIC base address
#define USART_PIO			AT91C_BASE_PIOA	//PIO controller that muxes USART pins
#define PIO_AB_SEL_REG		PIO_ASR			//select PIO_ASR or PIO_BSR
#define USART_RX_MASK		AT91C_PA0_RXD0	//PIO pin multiplexed with RXD line
#define USART_TX_MASK		AT91C_PA1_TXD0	//PIO pin multiplexed with TXD line


static volatile uint16_t txCounter;
static volatile char * txBuffer;

//prototype for USART Interrupt Service Routine
void usartIsr(void);

//send text via usart
void usartIrqSend(char *buffer, uint16_t len)
{
	//disable TX interrupt (if previous tx is already running)
	USART_BASE->US_IDR = AT91C_US_TXRDY;
	
	//reinitialize counter used in ISR
	txCounter = len;
	
	//set buffer pointer to the new buffer
	txBuffer = buffer;
	
	//enable tx interrupt on TXRDY flag
	USART_BASE->US_IER = AT91C_US_TXRDY;
}


//initialization func
void usartIrqInit(void)
{
	//enable USART clock
	AT91C_BASE_PMC->PMC_PCER = (1<<USART_ID);
	
	//disable PIO control on USART RX/TX pins
	USART_PIO->PIO_PDR = USART_RX_MASK | USART_TX_MASK;
	
	//select peripheral A or B register for USART pins
	USART_PIO->PIO_AB_SEL_REG = USART_RX_MASK | USART_TX_MASK;
	
	//reset and disable transmitter and receiver
	USART_BASE->US_CR = AT91C_US_RSTRX | AT91C_US_RSTTX | AT91C_US_RXDIS | AT91C_US_TXDIS;
	
	//set baud rate
	USART_BASE->US_BRGR = AT91F_US_Baudrate(MCK,USART_BAUDRATE);
	
	//set mode of operation
	USART_BASE->US_MR = AT91C_US_PAR_NONE | AT91C_US_CHRL_8_BITS;
	
	//configure AIC to work with usart
	const uint8_t USART_IRQ_PRIORITY = 5;
	AT91F_AIC_ConfigureIt ( AIC_BASE , USART_ID , USART_IRQ_PRIORITY , AT91C_AIC_SRCTYPE_INT_HIGH_LEVEL , usartIsr );
	//edge sensitive mode is not recommended
	
	//enable USART interrupt source in AIC
	AT91F_AIC_EnableIt ( AT91C_BASE_AIC , USART_ID );
	
	//enable interrupt sources form USART (select bits that will generate an interrupt)
	USART_BASE->US_IER = AT91C_US_RXRDY;
	
	//enable transmitter and receiver
	USART_BASE->US_CR = AT91C_US_RXEN | AT91C_US_TXEN;
}


//usart interrupt service routine
void usartIsr(void)
{
	
	//copy USART status reg to a variable
	uint32_t status;
	status = USART_BASE->US_CSR;
	
	//check if a char was rxed
	if(status & AT91C_US_RXRDY)
	{
		uint8_t rxedChar = USART_BASE->US_RHR;
		debug_msg("USART IRQ: rxed char = "); debug_chr(rxedChar);
		lcdLocate(0,0); lcdText("USART IRQ: chr="); lcdChr(rxedChar);
		return;
	}
	
	if(txCounter)
	{
		//check if ready to transmit
		if(status & AT91C_US_TXRDY)
		{
			txCounter--;
			USART_BASE->US_THR = *txBuffer;		//send next char
			txBuffer++;		//point the next char
		}
	}
	else
	{
		//buffer is sent
		//disable interrupt on TXRDY flag
		USART_BASE->US_IDR = AT91C_US_TXRDY;	
		debug_msg("tx complete!");
		lcdLocate(0,0); lcdText("TX complete!"); lcdText("    ");
	}
	
}


